//  Copyright (c) 2005, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <semaphore.h>


typedef struct _Message {
    int destination;
    int source;
    int sequenceNumber;
} Message;

typedef struct option cliOption;
typedef struct sigaction signalAction;


static inline int randomInteger(int min, int max) {
    return (min + (random() % (max - min + 1)));
}


int main(int argc, char* const argv[]) {
    const int defaultNumberOfProcesses = 4, defaultNumberOfMessages = 20, defaultMinimumWaitTime = 0, defaultMaximumWaitTime = 1000000;
    int numberOfProcesses = defaultNumberOfProcesses, numberOfMessages = defaultNumberOfMessages, minimumWaitTime = defaultMinimumWaitTime, maximumWaitTime = defaultMaximumWaitTime;
    int *pipes = NULL;
    cliOption longopts[] = { { "logfile", required_argument, NULL, 'f' },
                             { "help", no_argument, NULL, 'h' },
                             { "messages", required_argument, NULL, 'm' },
                             { "processes", required_argument, NULL, 'p' },
                             { "minwaittime", required_argument, NULL, '-' },
                             { "maxwaittime", required_argument, NULL, '+' },
                             { NULL, 0, NULL, 0 } };
    signalAction signal;
    int opt, i, err, thisProcess = 0;
    FILE *log = stdout;
    fd_set selectors;
    struct timeval timeout;
    Message message;
#ifdef LOCKED_LOGGING
#ifdef __linux__
    int flock(int fd, int operation);
#define LOG(message, ...) flock(fileno(log), LOCK_EX); fprintf(log, message, ## __VA_ARGS__); fflush(log); flock(fileno(log), LOCK_UN);
#else /* __linux__ */
    sem_t *sem;
    
    sem = sem_open("suma", O_CREAT | O_EXCL);
    
    if (NULL == sem) {
        fprintf(stderr, "Unable to initialise semaphore, error #%d (%s).\n", errno, strerror(errno));
        exit(errno);
    } else {
        sem_post(sem);
    }
    
    #define LOG(message, ...) sem_wait(sem); fprintf(log, message, ## __VA_ARGS__); fflush(log); sem_post(sem);
#endif /* __linux__ */
#else /* LOCKED_LOGGING */
    #define LOG(message, ...) fprintf(log, message, ## __VA_ARGS__); fflush(log);
#endif /* LOCKED_LOGGING */
    
    signal.sa_flags = 0;
    signal.sa_handler = SIG_IGN;
    sigemptyset(&(signal.sa_mask));
    
    err = sigaction(SIGPIPE, &signal, NULL);
    
    if (0 != err) {
        fprintf(stderr, "Unable to install signal handler, error #%d (%s).\n", errno, strerror(errno));
        exit(errno);
    }
    
    // Parse an CLI arguments
    
    while (-1 != (opt = getopt_long(argc, argv, "f:hm:p:", longopts, NULL))) {
        switch (opt) {
            case 'f': // Log file
                log = fopen(optarg, "w");
                
                if (NULL == log) {
                    fprintf(stderr, "Unable to open logfile \"%s\"; error #%d (%s).\n", optarg, errno, strerror(errno));
                    exit(errno);
                }
                
                break;
            case 'h': // Help/usage info
                printf("Usage: %s [OPTIONS]\n\n..where OPTIONS may be any of:\n\n\t--logfile/-f\tTakes a path as it's parameter, indicating the file to log events to.  Defaults to standard output.\n\t--help/-h\tPrints this usage information.\n\t--messages/-m\tTakes an integer parameter which is the number of messages each process should send.  Defaults to %d.\n\t--minwaittime\tTake an integer parameter which is the minimum number of microseconds to wait between messages (per individual client).  Defaults to %d.\n\t--maxwaittime\tTakes an integer parameter which is the maximum number of microseconds to wait between messages (per individual client).  Defaults to %d.\n\t--processes/-p\tTakes an integer parameter which is the number of processes to spawn for messaging.  Defaults to %d.\n", argv[0], defaultNumberOfMessages, defaultMinimumWaitTime, defaultMaximumWaitTime, defaultNumberOfProcesses);
                
                exit(0);
                
                break;
            case 'm': // Number of messages
                {
                    char *end;
                
                    numberOfMessages = strtol(optarg, &end, 0);
                    
                    if ((NULL == end) || (0 != *end) || (0 >= numberOfMessages)) {
                        fprintf(stderr, "Invalid parameter to the --messages/-m argument, \"%s\".  Expected a positive integer greater than 0.\n", optarg);
                        exit(EINVAL);
                    }
                }
                
                break;
            case 'p': // Number of processes
                {
                    char *end;
                
                    numberOfProcesses = strtol(optarg, &end, 0);
                
                    if ((NULL == end) || (0 != *end) || (1 >= numberOfProcesses)) {
                        fprintf(stderr, "Invalid parameter to the --processes/-p argument, \"%s\".  Expected a positive integer greater than 1.\n", optarg);
                        exit(EINVAL);
                    }
                }
                
                break;
            case '-': // Minimum wait time
                {
                    char *end;
                
                    minimumWaitTime = strtol(optarg, &end, 0);
                
                    if ((NULL == end) || (0 != *end) || (0 > minimumWaitTime)) {
                        fprintf(stderr, "Invalid parameter to the --minwaittime argument, \"%s\".  Expected a positive integer greater than or equal to 0.\n", optarg);
                        exit(EINVAL);
                    }
                }
                
                break;
            case '+': // Maximum wait time
                {
                    char *end;
                
                    maximumWaitTime = strtol(optarg, &end, 0);
                
                    if ((NULL == end) || (0 != *end) || (0 >= maximumWaitTime)) {
                        fprintf(stderr, "Invalid parameter to the --maxwaittime argument, \"%s\".  Expected a positive integer greater than 0.\n", optarg);
                        exit(EINVAL);
                    }
                }
                
                break;
            default:
                fprintf(stderr, "Unknown option.  See \"%s -h\" for usage information.\n", argv[0]);
                exit(EINVAL);
        }
    }
    
    if (minimumWaitTime > maximumWaitTime) {
        fprintf(stderr, "The maximum wait time must be greater than the minimum wait time (%d vs %d, respectively).\n", maximumWaitTime, minimumWaitTime);
        exit(EINVAL);
    }
    
    // Allocate the pipes we'll use for communication.
    
    pipes = malloc(sizeof(int) * 2 * numberOfProcesses);
    
    if (NULL == pipes) {
        fprintf(stderr, "Unable to allocate memory for pipes.\n");
        exit(ENOMEM);
    }
    
    for (i = 0; i < numberOfProcesses; ++i) {
        if (0 != pipe(&(pipes[2 * i]))) {
            fprintf(stderr, "Unable to create pipe pair #%d, error #%d (%s).\n", i, errno, strerror(errno));
            exit(errno);
        }
    }
    
    // Fork off our numerous processes
    
    for (i = 1; i < numberOfProcesses; ++i) {
        err = fork();
        
        if (-1 == err) {
            fprintf(stderr, "Unable to fork process #%d, error #%d (%s).\n", i, errno, strerror(errno));
            exit(errno);
        } else if (0 == err) { // We're the child, so exit this loop, after setting or process number.
            thisProcess = i;
            break;
        } // Else we're the parent; continue on.
    }
    
    // Now go through and close the read side of every pipe except our own, for which we close the write side.  This is not strictly necessary, but is a good idea as a way of conserving resources, and ensuring any mistakes in the code won't be able to send or receive messages from semantically invalid pipes.
    
    for (i = 0; i < numberOfProcesses; ++i) {
        if (thisProcess != i) {
            close(pipes[2 * i]);
            pipes[2 * i] = -1;
        } else {
            close(pipes[(2 * i) + 1]);
            pipes[(2 * i) + 1] = -1;
        }
    }
    
    // Now we can send our messages.
    
    for (i = 0; i < numberOfMessages; /* nothing */) {
        FD_ZERO(&selectors);
        FD_SET(pipes[2 * thisProcess], &selectors);
        
        timeout.tv_sec = 0;
        timeout.tv_usec = randomInteger(minimumWaitTime, maximumWaitTime);
        
        err = select(pipes[2 * thisProcess] + 1, &selectors, NULL, NULL, &timeout);
        
        if (-1 == err) {
            fprintf(stderr, "Unable to select pipe, error #%d (%s).\n", errno, strerror(errno));
            exit(errno);
        } else if (0 == err) { // Time to send our own message
            message.source = thisProcess;
            
            do {
                message.destination = randomInteger(0, numberOfProcesses - 1);
            } while (-1 == pipes[(2 * message.destination) + 1]);
            
            message.sequenceNumber = i;
            
            LOG("%4d - Sending message to %d with sequence number %d.\n", thisProcess, message.destination, message.sequenceNumber);
            
            err = write(pipes[(2 * message.destination) + 1], (const void*)&message, sizeof(message));
                        
            if (-1 == err) {
                if (EPIPE == errno) {
                    int j, allDead = 1;
                    
                    LOG("%4d - Noticed process %d is gone.\n", thisProcess, message.destination);
                    close(pipes[(2 * message.destination) + 1]);
                    pipes[(2 * message.destination) + 1] = -1;
                    
                    // This is probably unnecessary - on the next call to select we'll notice our read pipe has closed anyway.
                    
                    for (j = 0; (1 == allDead) && (j < numberOfProcesses); ++j) {
                        if (-1 != pipes[(2 * j) + 1]) {
                            allDead = 0;
                        }
                    }
                    
                    if (1 == allDead) {
                        LOG("%4d - All other processes appear to be gone.  Exiting with %d message%s unsent.\n", thisProcess, (numberOfMessages - i), ((1 == (numberOfMessages - i)) ? "" : "s"));
                        exit(0);
                    }
                } else {
                    fprintf(stderr, "Unable to write data to process #%d, error #%d (%s).\n", message.destination, errno, strerror(errno));
                    exit(EIO);
                }
            } else if (sizeof(message) != err) {
                fprintf(stderr, "Unable to write the entire message to process #%d.\n", message.destination);
                exit(EIO);
            } else {
                ++i;
            }
        } else { // We've got a message
            err = read(pipes[2 * thisProcess], (void*)&message, sizeof(message));
            
            if (-1 == err) {
                fprintf(stderr, "Unable to read from my pipe, despite select() saying it's got data - error #%d (%s).\n", errno, strerror(errno));
                exit(errno);
            } else if (0 == err) {
                LOG("%4d - All other processes appear to have finished; exiting with %d message%s unsent.\n", thisProcess, (numberOfMessages - i), ((1 == (numberOfMessages - i)) ? "" : "s"));
                exit(0);
            } else if (sizeof(message) != err) {
                fprintf(stderr, "Unable to read an entire message from the pipe... could buffer it, but... so... lazy.\n");
                exit(EIO);
            } else {
                LOG("%4d - Received message destined for %d, from %d, with sequence number %d.\n", thisProcess, message.destination, message.source, message.sequenceNumber);
            }
        }
    }
    
    LOG("%4d - All messages sent, exiting.\n", thisProcess);
    
    return 0;
}
